<?php
declare(strict_types=1);
namespace Flownative\GraphQL;

use GraphQL\Error\DebugFlag;
use GraphQL\Error\SyntaxError;
use GraphQL\Language\Parser;
use GraphQL\Server\ServerConfig;
use GraphQL\Server\StandardServer;
use GraphQL\Type\Schema;
use GraphQL\Utils\BuildSchema;
use GuzzleHttp\Psr7\Response;
use Neos\Cache\Exception as CacheException;
use Neos\Cache\Frontend\VariableFrontend;
use Neos\Flow\Annotations\CompileStatic;
use Neos\Flow\Annotations\Inject;
use Neos\Flow\Annotations\InjectConfiguration;
use Neos\Flow\Mvc\ActionRequest;
use Neos\Flow\ObjectManagement\ObjectManagerInterface;
use Neos\Flow\Reflection\ReflectionService;
use Neos\Flow\Security\Context as SecurityContext;
use Neos\Http\Factories\StreamFactory;
use Neos\Utility\Files;
use Psr\Http\Message\ResponseInterface;
use Psr\Http\Message\ServerRequestInterface;
use Psr\Http\Server\MiddlewareInterface;
use Psr\Http\Server\RequestHandlerInterface;


/**
 * An HTTP middleware which handles requests directed to GraphQL endpoints
 */
final class Middleware implements MiddlewareInterface
{
    /**
     * @InjectConfiguration
     * @var array
     */
    protected $settings;

    /**
     * @Inject
     * @var VariableFrontend
     */
    protected $schemaCache;

    /**
     * @Inject
     * @var ObjectManagerInterface
     */
    protected $objectManager;

    /**
     * @Inject
     * @var StreamFactory
     */
    protected $streamFactory;

    /**
     * @Inject(lazy=false)
     * @var SecurityContext
     */
    protected $securityContext;

    /**
     * @param ServerRequestInterface $request
     * @param RequestHandlerInterface $handler
     * @return ResponseInterface
     */
    public function process(ServerRequestInterface $request, RequestHandlerInterface $handler): ResponseInterface
    {
        $requestPath = $request->getUri()->getPath();
        foreach (self::getEndpointImplementations($this->objectManager) as $endpointClassName) {
            if ($requestPath === $endpointClassName::getPath()) {
                break;
            }
            unset($endpointClassName);
        }

        if (!isset($endpointClassName) || !in_array($request->getMethod(), ['POST', 'OPTIONS'])) {
            return $handler->handle($request);
        }

        if ($request->getMethod() === 'OPTIONS') {
            return new Response(200, ['Allow' => 'GET, POST, OPTIONS', 'Access-Control-Allow-Methods' => 'GET, POST, OPTIONS']);
        }

        $this->securityContext->setRequest(ActionRequest::fromHttpRequest($request));

        $endpoint = $this->objectManager->get($endpointClassName);
        assert($endpoint instanceof EndpointInterface);

        try {
            $schema = $this->getSchema($endpoint);
        } catch (SyntaxError | CacheException $e) {
            return new Response(500, ['Content-Type' => 'application/json'], sprintf('{error:"Failed processing schema from %s: %s"}', $endpoint->getSchemaUri(), $e->getMessage()));
        }

        $config = ServerConfig::create()
            ->setSchema($schema)
            ->setFieldResolver($endpoint)
            ->setDebugFlag(self::parseDebugOptions($this->settings['debug']))
        ;

        $server = new StandardServer($config);

        try {
            $request = $request->withParsedBody(json_decode($request->getBody()->getContents(), true, 512, JSON_THROW_ON_ERROR));
        } catch (\JsonException $e) {
            return new Response(400, ['Content-Type' => 'application/json'], '{error:"Failed decoding request body"}');
        }

        $bodyStream = $this->streamFactory->createStream();
        $response = $server->processPsrRequest($request, new Response(), $bodyStream);
        $bodyStream->rewind();

        return $response;
    }

    /**
     * @param array $options
     * @return int
     */
    private static function parseDebugOptions(array $options): int
    {
        $flag = DebugFlag::NONE;
        if ($options['includeDebugMessage'] === true) {
            $flag |= DebugFlag::INCLUDE_DEBUG_MESSAGE;
        }
        if ($options['includeTrace'] === true) {
            $flag |= DebugFlag::INCLUDE_TRACE;
        }
        return $flag;
    }

    /**
     * @param EndpointInterface $endpoint
     * @return Schema
     * @throws CacheException
     * @throws SyntaxError
     */
    private function getSchema(EndpointInterface $endpoint): Schema
    {
        $schemaPathAndFilename = $endpoint->getSchemaUri();
        if ($this->settings['enableSchemaCache'] === true) {
            $cacheKey = sha1($schemaPathAndFilename);
            if ($this->schemaCache->has($cacheKey)) {
                $documentNode = $this->schemaCache->get($cacheKey);
            } else {
                $documentNode = Parser::parse(Files::getFileContents($schemaPathAndFilename));
                $this->schemaCache->set($cacheKey, $documentNode);
            }
        } else {
            $documentNode = Parser::parse(Files::getFileContents($schemaPathAndFilename));
        }

        return BuildSchema::build($documentNode, $endpoint->getTypeConfigDecorator());
    }

    /**
     * @param ObjectManagerInterface $objectManager
     * @return array
     * @CompileStatic
     */
    protected static function getEndpointImplementations(ObjectManagerInterface $objectManager): array
    {
        $reflectionService = $objectManager->get(ReflectionService::class);
        return $reflectionService->getAllImplementationClassNamesForInterface(EndpointInterface::class);
    }
}
